/*
 * Copyright (c) 1996 Portland State University 
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by Portland State University.
 * 4. The name of Portland State University may not be used to endorse 
 *    or promote products derived from this software without specific 
 *    prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR/S ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR/S BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*
 * getmac.c
 *
 * function to take ifname as input and return mac address
 */

#include <sys/param.h>
#include <sys/socket.h>
#include <sys/ioctl.h>

#include <net/if.h>
#include <net/if_dl.h>
#include <netinet/in.h>
#include <netinet/in_var.h>
#include <arpa/inet.h>

#include <netns/ns.h>
#include <netns/ns_if.h>
#include <netdb.h>

#include <sys/protosw.h>

#include <ctype.h>
#include <err.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <nlist.h>
#include <kvm.h>
#include <fcntl.h>

#define MACMSIZ 30

kvm_t 	*kvmd;

static int kread(u_long addr, char *buf, int size);

/*
 * given an interface name, return a mac address
 *
 * NULL: on error
 */
char *
getmac(ifname)
	char ifname[];
{
	u_char *mac_addr;
	struct nlist nl[] = { { "_ifnet" } , "" };
	u_long addr, addr2;
	struct ifnet ifnet;
	union {
		struct ifaddr ifa;
		struct in_ifaddr in;
		struct ns_ifaddr ns;
	} ifaddr;
	char *cp=NULL;
	struct sockaddr *sa;
	struct sockaddr_dl *sdl;
	int n,m;
	char ifacename[IFNAMSIZ];

	mac_addr=(char *)malloc(MACMSIZ*sizeof(char));
	if (mac_addr == NULL)
		return(NULL);

	/*
	 * If we fail here it probably means we don't have permission to
	 * read /dev/kmem. Best to just silently bail out. If we have
	 * an error *after* we succeed in opening /dev/kmem, then we
	 * should report it.
	 */
	if ((kvmd = kvm_open(NULL,NULL,NULL,O_RDONLY,NULL)) == NULL)
		return(NULL);
	if (kvm_nlist(kvmd, nl) < 0 || nl[0].n_type == 0) {
#ifdef TESTMAIN
		perror("ifconfig: kvm_nlist()");
#else
		logperror("ifconfig: kvm_nlist()");
#endif
		return(NULL);
	}
	if (kread(nl[0].n_value, (char *)&addr, sizeof(addr))) {
#ifdef TESTMAIN
		perror("_ifnet");
#else
		logperror("_ifnet");
#endif
		return(NULL);
	}
	addr2 = 0;
	while (addr || addr2) {
		if (addr2 == 0) {
                        if (kread(addr, (char *)&ifnet, sizeof ifnet) ||
                           kread((u_long)ifnet.if_name, ifacename, IFNAMSIZ)){
#ifdef TESTMAIN
					perror("ifconfig: kvm_read()");
#else
					logperror("ifconfig: kvm_read()");
#endif
					return(NULL);
			}
			addr = (u_long)ifnet.if_next;
			addr2 = (u_long)ifnet.if_addrlist;
		}
		if (kread(addr2, (char *)&ifaddr, sizeof ifaddr)) {
			addr2 = 0;
                        continue;
                }
		sprintf(ifacename,"%s%d",ifacename, ifnet.if_unit);
		if (!strncmp(ifname, ifacename, strlen(ifname))) {
			int i;
			char *addr=mac_addr;
#define CP(x) ((char *)(x))
			cp = (CP(ifaddr.ifa.ifa_addr) - CP(addr2)) +
				CP(&ifaddr); sa = (struct sockaddr *)cp;
			if (sa->sa_family == AF_LINK) {
				sdl = (struct sockaddr_dl *)sa;
				cp = (char *)LLADDR(sdl);
				break;
			}
		}
		addr2 = (u_long)ifaddr.ifa.ifa_next;
	}
	kvm_close(kvmd);
	return(cp);
}

static int
kread(u_long addr, char *buf, int size)
{

        if (kvm_read(kvmd, addr, buf, size) != size)
                return (-1);
        return (0);
}

#ifdef TESTMAIN
main(argc, argv)
	int argc;
	char *argv[];
{
	int i;
	u_char *mac_addr;
	if (argc<2) {
		fprintf(stderr, "usage: ether if_name\n");
		exit(1);
	}
	if (mac_addr=getmac(argv[1])) {
		printf("MAC addr for %s is ", argv[1]);
		for (i=0;i<6;i++) 
			printf("%02x%c", *mac_addr++ & 0xff, (i<5)?':':' ');
		printf("\n");
	}
	else printf("No mac addr for %s\n", argv[1]);
}
#endif

