/*
 * Copyright (c) 1996,1997 Portland State University 
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by Portland State University.
 * 4. The name of Portland State University may not be used to endorse 
 *    or promote products derived from this software without specific 
 *    prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR/S ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR/S BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <sys/param.h>
#include <sys/queue.h>
#include <sys/socket.h>
#include <sys/socketvar.h>
#include <sys/protosw.h>

#include <netsec/ipsec.h>

#include <unistd.h>
#include <netdb.h>
#include <stdio.h>
#include "netstat.h"


void
dump_ah(off, name)
	u_long off;
	char *name;
{
	struct ah_stat_table AH_stats;

	if (off == 0)
		return;

        kread(off, (char *)&AH_stats, sizeof (AH_stats));
        printf("%s:\n", name);

#define	p(f, m) printf(m, AH_stats.f, plural(AH_stats.f))

	p(ah_sa_unavailable, "\t%u unavailable  AH Security Associations\n");
	p(ah_sa_invalid,     "\t%u invalid      AH Security Associations\n");
	p(ah_sa_delayed,     "\t%u delayed      AH Security Associations\n");


	p(ah_sent, "\t%u successful outgoing authentication%s (packets sent)\n");
	p(ah_accepted, "\t%u successful incoming authentication%s (packets accepted)\n");
	p(ah_in_errs, "\t%u failed incoming authentication%s (packets dropped)\n");
	p(ah_out_errs, "\t%u outgoing authentication%s failed (packets dropped)\n");
	p(ah_missing, "\t%u packet%s needed and were missing AH\n");
	p(ah_spoof, "\t%u packet%s attempted an AH tunnel spoof attack\n");

#undef p

	return;
}


void
dump_esp(off, name)
	u_long off;
	char *name;
{
	struct esp_stat_table ESP_stats;

	if (off == 0)
		return;

        kread(off, (char *)&ESP_stats, sizeof (ESP_stats));
        printf("%s:\n", name);

#define	p(f, m) printf(m, ESP_stats.f, plural(ESP_stats.f))

	p(esp_sa_unavailable, "\t%u unavailable ESP Security Associations\n");
	p(esp_sa_invalid,     "\t%u invalid     ESP Security Associations\n");
	p(esp_sa_delayed,     "\t%u delayed     ESP Security Associations\n");

	p(esp_transport_sent, "\t%u successful ESP transport-mode encryptions (packets sent)\n");
	p(esp_tunnel_sent, "\t%u successful ESP tunnel-mode encryptions sent (packets sent)\n");
	p(esp_accepted, "\t%u successful ESP decryptions performed (packets accepted)\n");
	p(esp_trans_missing, "\t%u incoming packets needed and were missing transport-mode ESP\n");
	p(esp_tunnel_missing, "\t%u incoming packets needed and were missing tunnel-mode ESP\n");
	p(esp_in_errs, "\t%u failed ESP decryption%s (packets dropped)\n");
	p(esp_trans_out_errs, "\t%u outgoing ESP transport-mode encryption%s failed (packets dropped)\n");
	p(esp_tunnel_out_errs, "\t%u outgoing ESP tunnel-mode encryption%s failed (packets dropped)\n");
	p(esp_tunnel_spoof, "\t%u received packet%s attempted ESP tunnel spoof attack\n");

#undef p
}


void
dump_ipsecip(off, name)
	u_long off;
	char *name;
{
	struct ipsec_ipstat ipsec_ipstat;

	if (off == 0)
		return;

        kread(off, (char *)&ipsec_ipstat, sizeof (ipsec_ipstat));
        printf("%s:\n", name);


	printf("\tip output pkts - ipsec needed %d\n",
        	ipsec_ipstat.ipsec_needed);
	printf("\tip output pkts - ipsec NOT needed %d\n",
		ipsec_ipstat.ipsec_notneeded); 
	printf("\tip output pkts - esp transport needed %d\n",
		ipsec_ipstat.ipsec_esp_transport_needed);
	printf("\tip output pkts - esp transport failed %d\n",
		ipsec_ipstat.ipsec_esp_transport_failed);

	printf("\tip output pkts - ah transport needed %d\n",
		ipsec_ipstat.ipsec_ah_transport_needed);
	printf("\tip output pkts - ah transport failed %d\n",
		ipsec_ipstat.ipsec_ah_transport_failed);     

	printf("\tip output pkts - esp tunnel needed %d\n",
		ipsec_ipstat.ipsec_esp_tunnel_needed);       
	printf("\tip output pkts - esp tunnel failed %d\n",
		ipsec_ipstat.ipsec_esp_tunnel_failed);

/* addition for PSU adaptation of NRL IPSec for MobileIP */
	printf("\tip output pkts - esp route needed %d\n",
		ipsec_ipstat.ipsec_route_esp_needed);
	printf("\tip output pkts - esp route tunnel needed %d\n",
		ipsec_ipstat.ipsec_route_esp_tunnel_needed);
	printf("\tip output pkts - esp route failed %d\n",
		ipsec_ipstat.ipsec_route_esp_failed);
	printf("\tip output pkts - ah route needed %d\n",
		ipsec_ipstat.ipsec_route_ah_needed);
	printf("\tip output pkts - ah route tunnel needed %d\n",
		ipsec_ipstat.ipsec_route_ah_tunnel_needed);
	printf("\tip output pkts - ah route failed %d\n",
		ipsec_ipstat.ipsec_route_ah_failed);
/* end -- addition for PSU */

	printf("\tip output pkts - sa not ready %d\n",
		ipsec_ipstat.ipsec_sa_not_ready_yet);
	printf("\tip output pkts - sa pkt drop %d\n",
		ipsec_ipstat.ipsec_sa_drop); 
	printf("\tip output pkts - ipsec error (EIPSEC) %d\n",
		ipsec_ipstat.ipsec_error);

	return;
}


