/*
 * Copyright (c) 1984, 1993
 *	The Regents of the University of California.  All rights reserved.
 *
 * This code is derived from software contributed to Berkeley by
 * Sun Microsystems, Inc.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by the University of
 *	California, Berkeley and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ifndef lint
static char const copyright[] =
"@(#) Copyright (c) 1984, 1993\n\
	The Regents of the University of California.  All rights reserved.\n";
#endif /* not lint */

#ifndef lint
static char const sccsid[] = "@(#)from: arp.c	8.2 (Berkeley) 1/2/94";
static char const freebsdid[] = "$Id: arp.c,v 1.3 1997/06/10 16:49:56 trost Exp $";
#endif /* not lint */

/*
 * arp - display, set, and delete arp table entries
 */


#include <sys/param.h>
#include <sys/file.h>
#include <sys/socket.h>
#include <sys/sockio.h>
#include <sys/sysctl.h>
#include <sys/ioctl.h>

#include <net/if.h>
#include <net/if_dl.h>
#include <net/if_types.h>
#include <net/route.h>

#include <netinet/in.h>
#include <netinet/if_ether.h>

#include <arpa/inet.h>

#include <netdb.h>
#include <errno.h>
#include <nlist.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <strings.h>
#include <paths.h>

void dump(u_long addr);
#ifdef IPSEC
	int delete(int argc, char *argv[]);
#else
	int delete(char *host, char *info);
#endif
void ether_print(u_char *cp);
void usage(void);
int set(int argc, char **argv);
void get(char *host);
int file(char *name);
void getsocket(void);
int ether_aton(char *a, u_char *n);
int rtmsg(int cmd);
void quit(char *msg);
int get_ether_addr(u_long ipaddr, u_char *hwaddr);

extern int errno;
static int pid;
static int nflag;
static int s = -1;

#ifdef IPSEC
#define MAXARGS 16      /* maximum number of args passed to set() */
extern int optind;
#endif
 
int
main(argc, argv)
	int argc;
	char **argv;
{
	int ch;
#ifdef IPSEC
	char *hostnull[2];
	hostnull[1] = NULL;
#endif

	pid = getpid();
#ifdef IPSEC
	while ((ch = getopt(argc, argv, "andfsScz")) != EOF)
#else
	while ((ch = getopt(argc, argv, "andfsSz")) != EOF)
#endif
		switch((char)ch) {
		case 'a':
			dump(0);
			exit(0);
#ifdef IPSEC
		case 'c':
			if (argc != 11)
				usage();
			exit(set(argc-1, &argv[1]) ? 1 : 0);
		case 'd':
			if (argc < 3 || argc > 6)
				usage();
			delete(argc-2, &argv[2]);
#else
		case 'd':
			if (argc < 3 || argc > 4)
				usage();
			delete(argv[2], argv[3]);
#endif /* IPSEC */
			exit(0);
		case 'n':
			nflag = 1;
			continue;
#ifdef IPSEC
		case 'S':
			if (argc < 4)
				usage();
			hostnull[0] = argv[2];
			delete(1, hostnull);
			/* FALL THROUGH */
		case 's':
			if (argc < 4 || argc > MAXARGS)
				usage();
			exit(set(argc-1, &argv[1]) ? 1 : 0);
#else
		case 'S':
			if (argc < 4)
				usage();
			delete(argv[2], NULL);
			/* FALL THROUGH */
		case 's':
			if (argc < 4 || argc > 7)
				usage();
			exit(set(argc-2, &argv[2]) ? 1 : 0);
#endif /* IPSEC */
		case 'f' :
			if (argc != 3)
				usage();
			return (file(argv[2]));
		case 'z':
			exit(flush());
			break;
		case '?':
		default:
			usage();
		}
#ifdef IPSEC
        if (argc == optind)
                usage();
        get(argv[optind]);
#else
	if (argc != 2)
		usage();
	get(argv[1]);
#endif
	exit(0);
}

/*
 * Process a file to set standard arp entries
 */
int
file(char *name)
{
	FILE *fp;
	int i, retval;
#ifdef IPSEC
	char line[1024], arg[MAXARGS-1][50], *args[MAXARGS-1];
#else
	char line[100], arg[5][50], *args[5];
#endif

	if ((fp = fopen(name, "r")) == NULL) {
		fprintf(stderr, "arp: cannot open %s\n", name);
		exit(1);
	}
	args[0] = &arg[0][0];
	args[1] = &arg[1][0];
	args[2] = &arg[2][0];
	args[3] = &arg[3][0];
	args[4] = &arg[4][0];
#ifdef IPSEC
        for (i = 5; i < MAXARGS-2; i++)
        	args[i] = &arg[i][0];
#endif

	retval = 0;
	while(fgets(line, 100, fp) != NULL) {
#ifdef IPSEC
		/* NOTE: only the add command (-s) is allowed, though
		 * including all new SA options.
		 */
		i = sscanf(line, "%s %s %s %s %s %s %s %s %s %s %s %s %s %s",
		    arg[1], arg[2], arg[3], arg[4], arg[5], arg[6], arg[7],
		    arg[8], arg[9], arg[10], arg[11], arg[12], arg[13], arg[14]
		    );
		strcpy(arg[0], "-s");
		i += 1;
#else
		i = sscanf(line, "%s %s %s %s %s",
				arg[0], arg[1], arg[2], arg[3], arg[4]);
#endif
		if (i < 2) {
			fprintf(stderr, "arp: bad line: %s\n", line);
			retval = 1;
			continue;
		}
		if (set(i, args))
			retval = 1;
	}
	fclose(fp);
	return (retval);
}

void
getsocket(void)
{
	if (s < 0) {
		s = socket(PF_ROUTE, SOCK_RAW, 0);
		if (s < 0) {
			perror("arp: socket");
			exit(1);
		}
	}
}

#ifdef IPSEC
struct  sockaddr_ipsec tmp_m = { sizeof(tmp_m), AF_INET, }, ah_m, esp_m;
struct	sockaddr_in so_mask = {sizeof(so_mask), 0, 0, { 0xffffffff}};
#else
struct	sockaddr_in so_mask = {8, 0, 0, { 0xffffffff}};
#endif
struct	sockaddr_in ifaddr = {8, AF_INET,0};
struct	sockaddr_inarp blank_sin = {sizeof(blank_sin), AF_INET }, sin_m;
struct	sockaddr_dl blank_sdl = {sizeof(blank_sdl), AF_LINK }, sdl_m;

int	expire_time, flags, export_only, doing_proxy, found_entry;

struct	{
	struct	rt_msghdr m_rtm;
	char	m_space[512];
}	m_rtmsg;

/* -if flag */
int ifflag;
char *ifname;
int if_index;

/*
 * Set an individual arp entry
 */
int
set(int argc, char **argv)
{
	struct hostent *hp;
	register struct sockaddr_inarp *sin = &sin_m;
	register struct sockaddr_dl *sdl;
	register struct rt_msghdr *rtm = &(m_rtmsg.m_rtm);
	u_char *ea;
	int command = RTM_ADD;
#ifdef IPSEC
	char *cmd = argv[0], *host = argv[1], *eaddr = argv[2];
	int gotspi = 0, gotitsrc = 0, gotitdst = 0;
	int ipsecused = 0;
#else
	char *host = argv[0], *eaddr = argv[1];
#endif
        int nullflag = 0;
        int rc; 

	getsocket();

#ifdef IPSEC
	if (strcmp("-c", cmd) == 0) {
		command = RTM_CHANGE;
		/*
		 * no link-layer address on a route change -- which
		 * also means we don't need to skip argc and argv past it.
		 */
		eaddr = "NULL";
	}
	else {
		argc--;
		argv++;
	}
#endif /* IPSEC */

	argc -= 2;
	argv += 2;
	sdl_m = blank_sdl;
	sin_m = blank_sin;
	sin->sin_addr.s_addr = inet_addr(host);
	if (sin->sin_addr.s_addr == -1) {
		if (!(hp = gethostbyname(host))) {
			fprintf(stderr, "arp: %s: ", host);
			herror((char *)NULL);
			return (1);
		}
		bcopy((char *)hp->h_addr, (char *)&sin->sin_addr,
		    sizeof sin->sin_addr);
	}
	doing_proxy = flags = export_only = expire_time = 0;
	while (argc-- > 0) {
		if (strncmp(argv[0], "temp", 4) == 0) {
			struct timeval time;
			gettimeofday(&time, 0);
			expire_time = time.tv_sec + 20 * 60;
		}
		else if (strncmp(argv[0], "pub", 3) == 0) {
			flags |= RTF_ANNOUNCE;
			doing_proxy = SIN_PROXY;
		} else if (strncmp(argv[0], "trail", 5) == 0) {
			printf("%s: Sending trailers is no longer supported\n",
				host);
		}
		/* idea is -if ed0; i.e., pass in name of interface */
		else if (strncmp(argv[0], "-if", 3) == 0) {
			ifflag = 1;

			if (argc)
				ifname = argv[1];
			else
				quit("arp: option -if requires an argument.");

			/* if eaddr is NULL, we can setup a empty
			 * link addr and not provide a gateway
			 */
			if (strcmp(eaddr, "NULL") == 0 ||
					strcmp(eaddr, "null") == 0)
			{
				nullflag = 1;
			}
			rc = getIfs(ifname, &if_index, &ifaddr);
			if (rc <= 0) {
				fprintf(stderr, "arp: invalid interface %s\n",
					ifname);
				exit(1);
			}
			argc--; argv++;
		} 
#ifdef IPSEC
		else if (strncmp(argv[0], "-t", 2) == 0) {
			char *ch;
			if (argc) {
				ch = argv[1];
			} else {
				quit("arp: option -type requires an argument.");
			}
			if (strcmp(ch, "ah") == 0) {
				tmp_m.sipsec_type = RTSA_AH;
				flags |= RTF_AUTH;
			}
			else if (strcmp(ch, "esp") == 0) {
				tmp_m.sipsec_type = RTSA_ESP;
				flags |= RTF_CRYPT;
			}
		}
		else if (strncmp(argv[0], "-id", 3) == 0) {
			int base = 10;
			char *ch;
			if (argc) {
				ch = argv[1]+1;
			} else {
				quit("arp: option -id (spi) requires an argument.");
			}
			if ( *ch == 'x' || *ch == 'X' )	base = 16;
			tmp_m.sipsec_spi = strtoul(argv[1], NULL, base);
			gotspi = 1;
			argc--; argv++;
		}
		else if (strncmp(argv[0], "-itsrc", 6) == 0) {
			filladdrin(argv[0], argv[1], &tmp_m.sipsec_src);
			gotitsrc = 1;
			argc--; argv++;
		}
		else if (strncmp(argv[0], "-itdst", 6) == 0) {
			filladdrin(argv[0], argv[1], &tmp_m.sipsec_dst);
			gotitdst = 1;
			argc--; argv++;
		}
#endif /* IPSEC */
		argv++;
	}

#ifdef IPSEC
	ipsecused = (flags & (RTF_AUTH|RTF_CRYPT));
	if ((ipsecused || gotspi || gotitsrc || gotitdst)
	    && !(ipsecused && gotspi && gotitsrc && gotitdst))
	{
		fprintf(stderr, "arp: security tuple incomplete.\n");
		usage();
	}
#endif /* IPSEC */

	ea = (u_char *)LLADDR(&sdl_m);
	if (doing_proxy && !strcmp(eaddr, "auto")) {
		if (!get_ether_addr(sin->sin_addr.s_addr, ea)) {
			return (1);
		}
		sdl_m.sdl_alen = 6;
	} else {
		if (strcmp(eaddr, "NULL") != 0 && strcmp(eaddr, "null") != 0)
			if (ether_aton(eaddr, ea) == 0)
				sdl_m.sdl_alen = 6;
	}
tryagain:
	/* get info from LLINFO/per interface/clone route
	*/
	if (!ifflag && rtmsg(RTM_GET) < 0) {
		perror(host);
		return (1);
	}
	sin = (struct sockaddr_inarp *)(rtm + 1);
	sdl = (struct sockaddr_dl *)(sin->sin_len + (char *)sin);
	if (sin->sin_addr.s_addr == sin_m.sin_addr.s_addr) {
		if (sdl->sdl_family == AF_LINK &&
		    (rtm->rtm_flags & RTF_LLINFO) &&
		    !(rtm->rtm_flags & RTF_GATEWAY)) switch (sdl->sdl_type) {
		case IFT_ETHER: case IFT_FDDI: case IFT_ISO88023:
		case IFT_ISO88024: case IFT_ISO88025:
			goto overwrite;
		}
		if (doing_proxy == 0) {
			printf("set: can only proxy for %s\n", host);
			return (1);
		}
		if (sin_m.sin_other & SIN_PROXY) {
			printf("set: proxy entry exists for non 802 device\n");
			return(1);
		}
		sin_m.sin_other = SIN_PROXY;
		export_only = 1;
		goto tryagain;
	}
overwrite:
        if (sdl->sdl_family != AF_LINK) {
                if (!ifflag) {
                        printf("cannot intuit interface index and type for %s\n"
, host);
                        return (1);
                }
                sdl_m.sdl_type = AF_LINK;
                sdl_m.sdl_index = if_index;
                ea = (u_char *)LLADDR(&sdl_m);
                /* XXX. not sure we need to do this again
                */

                if (nullflag)
                        sdl_m.sdl_alen = 0;
                else if (ether_aton(eaddr, ea) == 0)
                        sdl_m.sdl_alen = 6;
        }
        else {
                sdl_m.sdl_type = sdl->sdl_type;
                sdl_m.sdl_index = sdl->sdl_index;
        }

	return (rtmsg(command));
}

/*
 * Display an individual arp entry
 */
void
get(char *host)
{
	struct hostent *hp;
	struct sockaddr_inarp *sin = &sin_m;

	sin_m = blank_sin;
	sin->sin_addr.s_addr = inet_addr(host);
	if (sin->sin_addr.s_addr == -1) {
		if (!(hp = gethostbyname(host))) {
			fprintf(stderr, "arp: %s: ", host);
			herror((char *)NULL);
			exit(1);
		}
		bcopy((char *)hp->h_addr, (char *)&sin->sin_addr,
		    sizeof sin->sin_addr);
	}
	dump(sin->sin_addr.s_addr);
	if (found_entry == 0) {
		printf("%s (%s) -- no entry\n",
		    host, inet_ntoa(sin->sin_addr));
		exit(1);
	}
}

/*
 * Delete an arp entry
 */
int
#ifdef IPSEC
delete(int argc, char *argv[])
#else
delete(char *host, char *info)
#endif
{
	struct hostent *hp;
	register struct sockaddr_inarp *sin = &sin_m;
	register struct rt_msghdr *rtm = &m_rtmsg.m_rtm;
	struct sockaddr_dl *sdl;
#ifdef IPSEC
	char *host = NULL;
	struct sockaddr_ipsec	*sipsec = NULL;
	int ipsecused = 0;

	while (argc > 0) {
		if (strncmp(*argv, "-t", 2) == 0) {
			argv++; argc--;
			if (argc == 0 || !(**argv == 'a' || **argv == 'e')) {
				fprintf(stderr, "arp: argument to -type must be \"ah\" or \"esp\".\n");
				usage();
			}
			if (**argv == 'a') {
				tmp_m.sipsec_type = RTSA_AH;
			}
			else {
				tmp_m.sipsec_type = RTSA_ESP;
			}
			ipsecused = 1;
		}
		else if (**argv == '-') {
			usage();
		}
		else if (host) {
			/* parse trailing keyword(s) */
			if (strncmp(*argv, "pro", 3) != 0) {
				usage();
			}
			export_only = 1;
		}
		else {
			host = *argv;
		}
		argv++;
		argc--;
	}
#else
	if (info && strncmp(info, "pro", 3) )
		export_only = 1;
#endif /* IPSEC */

	getsocket();
	sin_m = blank_sin;
	sin->sin_addr.s_addr = inet_addr(host);
	if (sin->sin_addr.s_addr == -1) {
		if (!(hp = gethostbyname(host))) {
			fprintf(stderr, "arp: %s: ", host);
			herror((char *)NULL);
			return (1);
		}
		bcopy((char *)hp->h_addr, (char *)&sin->sin_addr,
		    sizeof sin->sin_addr);
	}
tryagain:
	if (rtmsg(RTM_GET) < 0) {
		perror(host);
		goto clean_rtmsg;
	}
	sin = (struct sockaddr_inarp *)(rtm + 1);
	sdl = (struct sockaddr_dl *)(sin->sin_len + (char *)sin);
	if (sin->sin_addr.s_addr == sin_m.sin_addr.s_addr) {
		if (sdl->sdl_family == AF_LINK &&
		    (rtm->rtm_flags & RTF_LLINFO) &&
		    !(rtm->rtm_flags & RTF_GATEWAY)) switch (sdl->sdl_type) {
		case IFT_ETHER: case IFT_FDDI: case IFT_ISO88023:
		case IFT_ISO88024: case IFT_ISO88025:
			goto delete;
		}
	}
	if (sin_m.sin_other & SIN_PROXY) {
		fprintf(stderr, "delete: can't locate %s\n",host);
		return (1);
	} else {
		sin_m.sin_other = SIN_PROXY;
		goto tryagain;
	}
delete:
	if (sdl->sdl_family != AF_LINK) {
		printf("cannot locate %s\n", host);
		return (1);
	}

#ifdef IPSEC
	/* build an ipsec "sockaddr" if necessary; clear any ipsec
	 * flags that might have been picked up by the GET above.
	 */
	rtm->rtm_flags &= ~(RTF_AUTH|RTF_CRYPT);
	rtm->rtm_addrs &= ~(RTA_AH|RTA_ESP);

	if (ipsecused) {
		if (tmp_m.sipsec_type == RTSA_AH) {
			rtm->rtm_flags |= RTF_AUTH;
			rtm->rtm_addrs |= RTA_AH;
		}
		else if (tmp_m.sipsec_type == RTSA_ESP) {
			rtm->rtm_flags |= RTF_CRYPT;
			rtm->rtm_addrs |= RTA_ESP;
		}
		else {
			quit("arp: (internal error) unrecognized route SA type.");
		}
		sipsec = (struct sockaddr_ipsec *)(sdl->sdl_len + (char *)sdl);
		bcopy(&tmp_m, sipsec, sizeof(struct sockaddr_ipsec));
		rtm->rtm_msglen += sizeof(struct sockaddr_ipsec);
	}
#endif /* IPSEC */

	if (rtmsg(RTM_DELETE) == 0) {
#ifdef IPSEC
		if (sipsec) {
			printf("%s (%s) deleted SA for %s.\n",
				host, inet_ntoa(sin->sin_addr),
				(tmp_m.sipsec_type==RTSA_AH)?"AH":"ESP");
			return 1;
		}
#endif
		printf("%s (%s) deleted\n", host, inet_ntoa(sin->sin_addr));
	}

clean_rtmsg:
	/* bzero m_rtmsg in case set is called immediately after and the -if
	 * option prevents a RTM_GET which would otherwise rewrite its
	 * contents.
	 */
	bzero(&m_rtmsg, sizeof(m_rtmsg));

	return (1);
}

/*
 * Dump the entire arp table
 */
void
dump(u_long addr)
{
	int mib[6];
	size_t needed;
	char *host, *malloc(), *lim, *buf, *next;
	struct rt_msghdr *rtm;
	struct sockaddr_inarp *sin;
	struct sockaddr_dl *sdl;
	extern int h_errno;
	struct hostent *hp;

	mib[0] = CTL_NET;
	mib[1] = PF_ROUTE;
	mib[2] = 0;
	mib[3] = AF_INET;
	mib[4] = NET_RT_FLAGS;
	mib[5] = RTF_LLINFO;
	if (sysctl(mib, 6, NULL, &needed, NULL, 0) < 0)
		quit("route-sysctl-estimate");
	if ((buf = malloc(needed)) == NULL)
		quit("malloc");
	if (sysctl(mib, 6, buf, &needed, NULL, 0) < 0)
		quit("actual retrieval of routing table");
	lim = buf + needed;
	for (next = buf; next < lim; next += rtm->rtm_msglen) {
		rtm = (struct rt_msghdr *)next;
		sin = (struct sockaddr_inarp *)(rtm + 1);
		sdl = (struct sockaddr_dl *)(sin + 1);
		if (addr) {
			if (addr != sin->sin_addr.s_addr)
				continue;
			found_entry = 1;
		}
		if (nflag == 0)
			hp = gethostbyaddr((caddr_t)&(sin->sin_addr),
			    sizeof sin->sin_addr, AF_INET);
		else
			hp = 0;
		if (hp)
			host = hp->h_name;
		else {
			host = "?";
			if (h_errno == TRY_AGAIN)
				nflag = 1;
		}
		printf("%s (%s) at ", host, inet_ntoa(sin->sin_addr));
		if (sdl->sdl_alen)
			ether_print(LLADDR(sdl));
		else
			printf("(incomplete)");
		if (rtm->rtm_rmx.rmx_expire == 0)
			printf(" permanent");
		if (sin->sin_other & SIN_PROXY)
			printf(" published (proxy only)");
		if (rtm->rtm_addrs & RTA_NETMASK) {
			sin = (struct sockaddr_inarp *)
				(sdl->sdl_len + (char *)sdl);
			if (sin->sin_addr.s_addr == 0xffffffff)
				printf(" published");
			if (sin->sin_len != 8)
				printf("(wierd)");
		}
		printf("\n");
	}
}

void
ether_print(u_char *cp)
{
	printf("%x:%x:%x:%x:%x:%x", cp[0], cp[1], cp[2], cp[3], cp[4], cp[5]);
}

int
ether_aton(char *a, u_char *n)
{
	int i, o[6];

	i = sscanf(a, "%x:%x:%x:%x:%x:%x", &o[0], &o[1], &o[2],
					   &o[3], &o[4], &o[5]);
	if (i != 6) {
		fprintf(stderr, "arp: invalid Ethernet address '%s'\n", a);
		return (1);
	}
	for (i=0; i<6; i++)
		n[i] = o[i];
	return (0);
}

void
usage(void)
{
#ifdef IPSEC                                    /* new usage statement */
#define p(s) printf("    arp "s);

        printf("usage:\n");

 p("[-n] hostname\n");
 p("[-n] -a [kernel] [kernel_memory]\n");
 p("-c hostname SA_tuple\n");
 p("-S hostname ether_addr | NULL [temp] [pub] [SA_tuple]\n");
 p("-s hostname ether_addr | NULL [temp] [pub] [-if <i/f_name>] [SA_tuple]\n");
 p("-d hostname [proxy] [-type (ah|esp)]\n");
 p("-f filename\n");
 p("-z\n");

        printf("\n");
        printf("\tSA_tuple = -type (ah|esp) -id <SPI> -itsrc <IPaddr> -itdst <IPaddr>\n");
        /* printf("\t<ether_addr> may be \"NULL\".\n\n"); */

#undef p

#else                                
	printf("usage: arp hostname\n");
	printf("       arp -a [kernel] [kernel_memory]\n");
	printf("       arp -d hostname\n");
	printf("       arp -s hostname ether_addr [temp] [pub]\n");
	printf("       arp -S hostname ether_addr [temp] [pub]\n");
	printf("       arp -f filename\n");
#endif
	exit(1);
}

int
rtmsg(int cmd)
{
	static int seq;
	int rlen;
	register struct rt_msghdr *rtm = &m_rtmsg.m_rtm;
	register char *cp = m_rtmsg.m_space;
	register int l;

	errno = 0;
	if (cmd == RTM_DELETE)
		goto doit;
	bzero((char *)&m_rtmsg, sizeof(m_rtmsg));
	rtm->rtm_flags = flags;
	rtm->rtm_version = RTM_VERSION;

	switch (cmd) {
	default:
		fprintf(stderr, "arp: internal wrong cmd\n");
		exit(1);
#ifdef IPSEC
	case RTM_CHANGE:
		if (expire_time || doing_proxy) {
			/* improper arguments given, but parsed by set() */
			usage();
		}
		rtm->rtm_flags |= (RTF_HOST|RTF_STATIC);
		if (ifflag)
			rtm->rtm_addrs |= RTA_IFA;
		rtm->rtm_addrs |= RTA_DST;
		break;
#endif
	case RTM_ADD:
		rtm->rtm_addrs |= RTA_GATEWAY;
		rtm->rtm_rmx.rmx_expire = expire_time;
		rtm->rtm_inits = RTV_EXPIRE;
		rtm->rtm_flags |= (RTF_HOST | RTF_STATIC);
		sin_m.sin_other = 0;
		if (doing_proxy) {
			if (export_only)
				sin_m.sin_other = SIN_PROXY;
			else {
				rtm->rtm_addrs |= RTA_NETMASK;
				rtm->rtm_flags &= ~RTF_HOST;
			}
		}
                if (ifflag)
                        rtm->rtm_addrs |= RTA_IFA; 
		/* FALLTHROUGH */
	case RTM_GET:
		rtm->rtm_addrs |= RTA_DST;
	}

#ifdef IPSEC
	if (rtm->rtm_flags & RTF_AUTH) {
		rtm->rtm_addrs |= RTA_AH;
		bcopy(&tmp_m, &ah_m, sizeof(tmp_m));
	}
	else if (rtm->rtm_flags & RTF_CRYPT) {
		rtm->rtm_addrs |= RTA_ESP;
		bcopy(&tmp_m, &esp_m, sizeof(tmp_m));
	}
#endif

#define NEXTADDR(w, s) \
	if (rtm->rtm_addrs & (w)) { \
		bcopy((char *)&s, cp, sizeof(s)); cp += sizeof(s);}

	NEXTADDR(RTA_DST, sin_m);
	NEXTADDR(RTA_GATEWAY, sdl_m);
	NEXTADDR(RTA_NETMASK, so_mask);
        NEXTADDR(RTA_IFA, ifaddr);
#ifdef IPSEC
        NEXTADDR(RTA_AH, ah_m);
        NEXTADDR(RTA_ESP, esp_m);
#endif /* IPSEC */

	rtm->rtm_msglen = cp - (char *)&m_rtmsg;
doit:
	l = rtm->rtm_msglen;
	rtm->rtm_seq = ++seq;
	rtm->rtm_type = cmd;
	if ((rlen = write(s, (char *)&m_rtmsg, l)) < 0) {
		if (errno != ESRCH || cmd != RTM_DELETE) {
			perror("writing to routing socket");
			return (-1);
		}
	}
	do {
		l = read(s, (char *)&m_rtmsg, sizeof(m_rtmsg));
	} while (l > 0 && (rtm->rtm_seq != seq || rtm->rtm_pid != pid));
	if (l < 0)
		(void) fprintf(stderr, "arp: read from routing socket: %s\n",
		    strerror(errno));
	return (0);
}

#ifdef IPSEC
filladdrin(option, host, addr)
char *option;
char *host;
u_long *addr;
{
        struct hostent *hp;

	if (!host) {
		char buf[1024];
		sprintf(buf, "arp: option %s requires an argument.", option);
		quit(buf);
	}

	*addr = inet_addr(host);
        if (*addr == 0xffffffff) {
                if (!(hp = gethostbyname(host))) {
                        fprintf(stderr, "arp: %s: ", host);
                        herror((char *)NULL);
                        return (1);
                }
                bcopy((char *)hp->h_addr, (char *)addr, sizeof(u_long));
        }
}
#endif /* IPSEC */

/*
 *  flush all arp entries.
 *  errors due to a GET or DELETE on the routing socket are counted
 *  	and returned.
 */
flush()
{
	int mib[6];
	size_t needed;
	char *malloc(), *lim, *buf, *next;
	struct rt_msghdr *rtm;
	int error = 0;

	mib[0] = CTL_NET;
	mib[1] = PF_ROUTE;
	mib[2] = 0;
	mib[3] = AF_INET;
	mib[4] = NET_RT_FLAGS;
	mib[5] = RTF_LLINFO;

	if (sysctl(mib, 6, NULL, &needed, NULL, 0) < 0)
		quit("route-sysctl-estimate (flush)");
	if ((buf = malloc(needed)) == NULL)
		quit("malloc (flush)");
	if (sysctl(mib, 6, buf, &needed, NULL, 0) < 0)
		quit("actual retrieval of routing table (flush)");
	getsocket();
	lim = buf + needed;
	for (next = buf; next < lim; next += rtm->rtm_msglen) {
		rtm = (struct rt_msghdr *) next;
		bcopy(&blank_sin, &sin_m, sizeof(struct sockaddr_inarp));
		bcopy( &((struct sockaddr_inarp *)(rtm + 1))->sin_addr.s_addr,
		       &(sin_m.sin_addr.s_addr),
		       sizeof(long));

		if (rtmsg(RTM_GET) < 0) {
			perror("arp: RTM_GET failed in flush()\n");
			error += 1;
			continue;
		}

#ifdef IPSEC
		/* turn off all ipsec flags to signify a full delete.
		 * NOTE: this turns off the bits but doesn't zero out
		 * the sockaddrs containing ipsec data which still get
		 * sent down the socket.
		 */
		m_rtmsg.m_rtm.rtm_flags &= ~(RTF_AUTH|RTF_CRYPT);
		m_rtmsg.m_rtm.rtm_addrs &= ~(RTA_AH|RTA_ESP);
#endif /* IPSEC */

		if (rtmsg(RTM_DELETE)) {
			perror("arp: RTM_DELETE failed in flush()\n");
			error += 1;
		}
	}

	return(error);
}

void
quit(char *msg)
{
	fprintf(stderr, "%s\n", msg);
	exit(1);
}

/*
 * get_ether_addr - get the hardware address of an interface on the
 * the same subnet as ipaddr.
 */
#define MAX_IFS		32

int
get_ether_addr(u_long ipaddr, u_char *hwaddr)
{
	struct ifreq *ifr, *ifend, *ifp;
	u_long ina, mask;
	struct sockaddr_dl *dla;
	struct ifreq ifreq;
	struct ifconf ifc;
	struct ifreq ifs[MAX_IFS];
	int s;

	s = socket(AF_INET, SOCK_DGRAM, 0);
	if (s < 0) {
		perror("socket");
		exit(1); 
	}

	ifc.ifc_len = sizeof(ifs);
	ifc.ifc_req = ifs;
	if (ioctl(s, SIOCGIFCONF, &ifc) < 0) {
		fprintf(stderr, "ioctl(SIOCGIFCONF): \n");
		close(s);
		return 0;
	}

	/*
	* Scan through looking for an interface with an Internet
	* address on the same subnet as `ipaddr'.
	*/
	ifend = (struct ifreq *) (ifc.ifc_buf + ifc.ifc_len);
	for (ifr = ifc.ifc_req; ifr < ifend; ) {
		if (ifr->ifr_addr.sa_family == AF_INET) {
			ina = ((struct sockaddr_in *) 
				&ifr->ifr_addr)->sin_addr.s_addr;
			strncpy(ifreq.ifr_name, ifr->ifr_name, 
				sizeof(ifreq.ifr_name));
			/*
			 * Check that the interface is up,
			 * and not point-to-point or loopback.
			 */
			if (ioctl(s, SIOCGIFFLAGS, &ifreq) < 0)
				continue;
			if ((ifreq.ifr_flags &
			     (IFF_UP|IFF_BROADCAST|IFF_POINTOPOINT|
					IFF_LOOPBACK|IFF_NOARP))
			     != (IFF_UP|IFF_BROADCAST))
				goto nextif;
			/*
			 * Get its netmask and check that it's on 
			 * the right subnet.
			 */
			if (ioctl(s, SIOCGIFNETMASK, &ifreq) < 0)
				continue;
			mask = ((struct sockaddr_in *)
				&ifreq.ifr_addr)->sin_addr.s_addr;
			if ((ipaddr & mask) != (ina & mask))
				goto nextif;
			break;
		}
nextif:
		ifr = (struct ifreq *) 
		    ((char *)&ifr->ifr_addr + ifr->ifr_addr.sa_len);
	}

	if (ifr >= ifend) {
		close(s);
		return 0;
	}

	/*
	* Now scan through again looking for a link-level address
	* for this interface.
	*/
	ifp = ifr;
	for (ifr = ifc.ifc_req; ifr < ifend; ) {
		if (strcmp(ifp->ifr_name, ifr->ifr_name) == 0
		    && ifr->ifr_addr.sa_family == AF_LINK) {
			/*
			 * Found the link-level address - copy it out
			 */
		 	dla = (struct sockaddr_dl *) &ifr->ifr_addr;
			memcpy(hwaddr,  LLADDR(dla), dla->sdl_alen);
			close (s);
			printf("using interface %s for proxy with address ",
				ifp->ifr_name);
			ether_print(hwaddr);
			printf("\n");
			return dla->sdl_alen;
		}
		ifr = (struct ifreq *) 
			((char *)&ifr->ifr_addr + ifr->ifr_addr.sa_len);
	}
	return 0;
}

